/*
   This file is part of the Aulib library.
   Name: afplayer.cpp
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "afplayer.h"

#ifdef AF
#include "error.h"
#include "audata.h"
#include <algorithm>
using namespace std;

////////////////////////////////////////////////////////////

float AFPlayer::BUFSIZE = 0.001;

////////////////////////////////////////////////////////////

AFPlayer::AFPlayer(int device, const string& server)
  : _server(server), _device(device), _gain(0), _delay(0.1),
    _aud(0), _time(0) {

  char* sname = const_cast<char*>(_server.c_str());
  if ((_aud = AFOpenAudioConn(sname)) == 0) {
    throw Exception("AFPlayer::AFPlayer()"
      " - No connection to audio server possible.");
  }
}

////////////////////////////////////////////////////////////

long int AFPlayer::calcLoops(const AudioData& audata, long int length) const {

  // number of bytes from the current position to the EOF
  long int remain = audata._dataLocation + audata._dataSize
    - audata._file.tellg();

  // ...and in samples
  long int samples = remain / (audata._bytesInSample * 
    audata._channelCount);

  return min(length, samples);
}

////////////////////////////////////////////////////////////

void AFPlayer::setStartingTime(AC ac) {

  ATime t2 = AFGetTime(ac);

  if (_time == 0) {
    _time = t2
     + static_cast<ATime>(_delay * ac->device->playSampleFreq);
  }
  else {
    if (_time < t2) {
      _time = t2;
    }
  }
}

////////////////////////////////////////////////////////////

void AFPlayer::play(const AudioData& audata, long int length) {
  AC ac = createACFor(audata);
  setStartingTime(ac);
  long int loops = calcLoops(audata, length);
  playFile(audata, ac, loops);
  deleteAC(ac);
}

////////////////////////////////////////////////////////////

void AFPlayer::play(const AudioData& audata) {
  // data size in bytes is always greather or equal to
  // the data size in samples, so play() will play audio
  // data to the end of file.
  play(audata, audata.dataSize());
}

////////////////////////////////////////////////////////////

void AFPlayer::playFile(const AudioData& audata, AC ac, long int loops) {

  const int UnitSize = calcUnitSize(audata);       // size of AF unit
  const int BufSize = calcPlayBufSize(audata, ac); // in seconds
  Byte buf[BufSize * UnitSize];                    // play buffer
  int nbytes;   // number of bytes read from file
  int units;    // number of units
  int nsamples; // number of samples in buffer
  ATime nact;   // AF timing

  // append samples to buffer and write buffer to device
  // each loop goes ahaed in 1 hsec steps.
  for(long int i = 0; i < loops; i++) {
    audata._file.read(buf, BufSize * UnitSize);
    if (audata._file.fail())
      throw Exception("AFPlayer::playFile() - error reading file.");
    nbytes = audata._file.gcount();
    units = nbytes / UnitSize;
    nsamples = units * SampsPerUnit(ac->device->playBufType);
    nact = AFPlaySamples(ac, _time, units * UnitSize, buf);
    _time += nsamples;
  }
}

////////////////////////////////////////////////////////////

AC AFPlayer::createACFor(const AudioData& audata) {
  ADevice usedDevice = findDeviceFor(audata);
  AFSetACAttributes attrib = getAttributesFor(audata);
  AC ac = createAC(audata, usedDevice, attrib);
  return ac;
}

////////////////////////////////////////////////////////////

ADevice AFPlayer::findDeviceFor(const AudioData& audata) const {
  if (_device < 0 || _device >= _aud->ndevices)
    return findDefaultDeviceFor(audata);
  else
    return static_cast<ADevice>(_device);
}

////////////////////////////////////////////////////////////

AFSetACAttributes AFPlayer::getAttributesFor(const AudioData& audata) const {

  AFSetACAttributes attrib;

#ifdef DAMN_INTEL_BYTE_ORDER
  attrib.endian = ALittleEndian;
#else
  attrib.endian = ABigEndian;
#endif

  attrib.preempt = Mix;
  attrib.start_timeout = 0;
  attrib.end_silence = 0;
  attrib.play_gain = _gain;
  attrib.rec_gain = 0;
  attrib.channels = audata.channelCount();
  attrib.type = audata.AFEncodeType();

  // check if AF supports encoding type
  if (attrib.type == UNKNOWN_ENCODETYPE ||
      AF_sample_sizes[attrib.type].samps_per_unit == 0)
    throw Exception("AFPlayer::getAttributesFor() - AF cannot play this encoding type.");

  return attrib;
}

////////////////////////////////////////////////////////////

AC AFPlayer::createAC(const AudioData& audata, ADevice device, AFSetACAttributes attrib) const {

  AC ac = AFCreateAC(_aud, device, (ACPlayGain | ACEncodingType
    | ACChannels | ACEndian), &attrib);

  // make sure confirm encoding type support
  AFSync(_aud, 0);
  
  // check if audio files's frequncy suits to device
  if (audata.samplingRate() > (ac->device->playSampleFreq + 100) ||
      audata.samplingRate() < (ac->device->playSampleFreq - 100))
    throw Exception ("Device cannot play this frequency.");

  // check if file's number of channels suits to device
  if (audata.channelCount() != ac->device->playNchannels)
    throw Exception("Device cannot play so much channels.");

  return ac;
}

////////////////////////////////////////////////////////////

ADevice AFPlayer::findDefaultDeviceFor(const AudioData& audata) const {

  AFDeviceDescriptor* aDev;
  int device;
  
  // using best fit for audio stream
  for(device = 0; device < _aud->ndevices; device++) {
    aDev = AAudioDeviceDescriptor(_aud, device);
    if (aDev->playNchannels == audata._channelCount &&
       (aDev->playSampleFreq < (audata._samplingRate + 100) ||
        aDev->playSampleFreq > (audata._samplingRate - 100)))
      return static_cast<ADevice>(device);
  }

  // no optimal device found: using first device not connected
  // to phone
  for(device = 0; device < _aud->ndevices; device++) {
    aDev = AAudioDeviceDescriptor(_aud, device);
    if (aDev->inputsFromPhone == 0 && aDev->outputsToPhone == 0)
      return static_cast<ADevice>(device);
  }

  throw Exception("AFPlayer::findDefaultDeviceFor() - No device found which can play this file.");

  return static_cast<ADevice>(device);
}

#endif

