/*
   This file is part of the Aulib library.
   Name: aucache.h
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef _AU_CACHE_H_
#define _AU_CACHE_H_

#include "autypes.h"
#include "austruct.h"
#include "auformat.h"

#ifdef OLDGNU
#include <iostream.h>
#else
#include <iostream>
#endif

using namespace std;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                    AudioCache

/**
  * The <tt>AudioCache</tt> class represents a vector of samples.
  * The data format is mono, lin16 (2 bytes per sample, linear
  * encoding). The sampling rate can be set only once when the cache
  * is created.
  * <p>
  * <tt>AudioCache</tt> is used as argument by the most of analyse
  * tools. <tt>AudioData</tt> uses it as a buffer to store parts 
  * of audio files.
  *
  * @see AudioData
  *
  * @author <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>, <a href = "mailto:tomczyk@chili.wifo.uni-mannheim.de">Peter Tomczyk</a>
  * @version $Id: aucache.h,v 1.10 1998/09/11 11:09:06 tomczyk Exp $
  */
class AudioCache : public Array<Sample> {

  public:

    /**
     * Creates new AudioCache that has 'size' samples. The
     * sampling rate of the stored signal is set to 'rate'.
     *
     * @param size cache size
     * @param rate sampling rate [Hz]
     * @exception Exception
     */
    explicit AudioCache(size_type size = 0, uint32 rate = 8000)
      : Array<Sample>(size), _samplingRate(rate) { }

    /// Creates an AudioCache that is a copy of 'c'.
    AudioCache(const AudioCache& c)
      : Array<Sample>(c), _samplingRate(c._samplingRate) { }

    /**
     * Assing the invoking AudioCache the contents of 'c'.
     * Throws an exception if the sampling rates are not
     * equal.
     *
     * @param c another AudioCache
     * @exception Exception
     */
    AudioCache& operator=(const AudioCache& c);

    /**
     * Returns true if the sampling rate and the contents of
     * the invoking cache are all equal to these of 'c'.
     */
    bool operator==(const AudioCache& c) const;

    /// Returns sampling rate of the stored signal.
    uint32 samplingRate() const { return _samplingRate; }

    /// Returns number of audio channels.
    uint32 channelCount() const { return 1; }

    /// Returns the size of a single sample in bytes.
    uint32 bytesInSample() const { return 2; }

    /// Returns data format of the stored signal.
    DataFormat::FormatCode dataFormat() const { 
      return DataFormat::lin16; }

  protected:

    /// Prints an info about the cache on the given stream.
    ostream& printOn(ostream& os) const;

    /// Sampling rate of the stored signal.
    const uint32 _samplingRate;
};

inline
bool AudioCache::operator==(const AudioCache& c) const {
  return (_samplingRate == c._samplingRate && Array<Sample>::operator==(c));
}

#endif

