/*
   This file is part of the Aulib library.
   Name: audata.h
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef _AU_DATA_H_
#define _AU_DATA_H_

#include "autypes.h"
#include "austruct.h"
#include "auformat.h"
#include "aucache.h"

#ifdef AF
#include <AF/AFlib.h>
#include <AF/AFUtils.h>
#endif

#include <string>
#include <cstddef>

#ifdef OLDGNU
#include <iostream.h>
#include <fstream.h>
#else
#include <iostream>
#include <fstream>
#endif

using namespace std;

class AFPlayer;
class ByteReader;
class ByteWriter;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       AudioData

/**
 * AudioData is the abstract base class of all classes which
 * represent audio files. It defines a set of methods to open
 * and close audio files, to load and store samples etc.
 * <P>
 * AudioData uses an AudioCache object to store parts of the
 * audio file so that processing is faster to be done.
 *
 * @see AudioCache
 *
 * @author <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>, <a href = "mailto:tomczyk@chili.wifo.uni-mannheim.de">Peter Tomczyk</a>
 * @version $Id: audata.h,v 1.19 1998/09/11 11:10:23 tomczyk Exp $
 */
class AudioData {
  friend class AFPlayer;
  friend ostream& operator<<(ostream& os, const AudioData& audata) {
    return audata.printOn(os);
  }
  public:
    enum SeekMode { BEG = 0, CUR = 1, END = 2 };

    explicit AudioData(const string& filename);
    virtual ~AudioData() = 0;

    virtual size_t load(size_t samples) const;
    virtual size_t loadSecs(size_t secs) const;
    virtual void seek(long int samples, SeekMode mode = BEG) const;
    virtual void seekSecs(int secs, SeekMode mode = BEG) const;
    virtual void store(ofstream& file) const;

    const string& filename() const { return _filename; }
    uint32 dataLocation() const { return _dataLocation; }
    uint32 dataSize() const { return _dataSize; }
    DataFormat::FormatCode dataFormat() const { return _dataFormat; }
    uint32 samplingRate() const { return _samplingRate; }
    uint32 bytesInSample() const { return _bytesInSample; }
    uint32 channelCount() const { return _channelCount; }
    const string& info() const { return _info; }
#ifdef AF
    AEncodeType AFEncodeType() const;
#endif

    size_t filepos() const;

    const AudioCache& cache() const { return *_cache; }
    AudioCache& cache() { return *_cache; }

//    virtual void playOn(AFPlayer& player, 
//                        const long length,
//                        const int gain = 0,
//                        const float delay = 0.1,
//                        const int device = -1) const = 0;

  protected:
    virtual ostream& printOn(ostream& os) const;

    mutable ifstream _file;
    string _filename;
    uint32 _dataLocation;
    uint32 _dataSize;
    DataFormat::FormatCode _dataFormat;
    uint32 _samplingRate;
    uint32 _bytesInSample;
    uint32 _channelCount;
    string _info;
    mutable AudioCache* _cache;

    ByteReader* _byteReader;
    ByteWriter* _byteWriter;

  private:
    AudioData(const AudioData&);
    AudioData& operator=(const AudioData&);
};

////////////////////////////////////////////////////////////

inline
size_t AudioData::filepos() const {
  return ((_file.tellg() - _dataLocation)
             / (_bytesInSample * _channelCount));
}

////////////////////////////////////////////////////////////

#ifdef AF
inline
AEncodeType AudioData::AFEncodeType() const { 
  return DataFormat::AFEncodeType(_dataFormat); }
#endif

////////////////////////////////////////////////////////////

inline
void AudioData::seekSecs(int secs, SeekMode mode) const {
  const long int samples = secs * _samplingRate * _channelCount;
  seek(samples, mode);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       AudioFile

/**
 * AudioFile represents a SUN-au file. It implements the interface
 * defined by AudioData.
 *
 * @author <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>, <a href = "mailto:tomczyk@chili.wifo.uni-mannheim.de">Peter Tomczyk</a>
 * @version $Id: audata.h,v 1.19 1998/09/11 11:10:23 tomczyk Exp $
 */
class AudioFile : public AudioData {
//  friend class AFPlayer;
  public:
    explicit AudioFile(const string& filename);
    ~AudioFile() { }

    uint32 headerID() const { return _headerID; }

//    void playOn(AFPlayer& player, const long length,
//                const int gain = 0, const float delay = 0.1, 
//                const int device = -1) const;

  protected:
    void readHeader();
    uint32 packBytes(Byte* bytes) const;
    Byte* swapInt32(Byte cp[4]) const;
    bool isSNDFile(Byte header[4]) const;

    uint32 _headerID;

  private:
    AudioFile(const AudioFile& aufile);
    AudioFile& operator=(const AudioFile&);

};

////////////////////////////////////////////////////////////

//void AudioFile::playOn(AFPlayer& player, const long length,
//                const int gain, const float delay, 
//                const int device) const {
//  player.play(*this, length, gain, delay, device);
//}

////////////////////////////////////////////////////////////

inline
uint32 AudioFile::packBytes(Byte* bytes) const {
#ifdef DAMN_INTEL_BYTE_ORDER
  return *(reinterpret_cast<uint32*>(swapInt32(bytes)));
#else
  return *(reinterpret_cast<uint32*>(bytes));
#endif
}

////////////////////////////////////////////////////////////

inline
bool AudioFile::isSNDFile(Byte header[4]) const {
  return ((header[0] == 0x2e) && (header[1] == 0x73)
       && (header[2] == 0x6e) && (header[3] == 0x64));
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       WaveFile

/**
 * WaveFile represents a WAV-file. It implements the interface
 * defined by AudioData.
 * <p>
 * The structure of an ordinary WAV-file is:
 * <pre>
 * 'RIFF' header              - 4 bytes
 * file size                  - 4 bytes
 * 'WAVE' header              - 4 bytes
 *
 * 'fmt ' header              - 4 bytes
 * length of the format chunk - 4 bytes (10 00 00 00H)
 * format tag                 - 2 bytes (must be 01 00H)
 * number of channels         - 2 bytes (1 = mono, 2 = stereo)
 * sampling rate              - 4 bytes
 * bytes per second           - 4 bytes
 * bytes per sample           - 2 bytes (1, 2 or 4)
 * bits per sample            - 2 bytes (8, 12 or 16)
 *
 * 'data' header              - 4 bytes
 * length of the data chunk   - 4 bytes
 * audio data...
 * </pre>
 * A WAV-file can contains of more than one data chunk.
 * WaveFile can only handle the first data chunk.
 *
 * @author <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>, <a href = "mailto:tomczyk@chili.wifo.uni-mannheim.de">Peter Tomczyk</a>
 * @version $Id: audata.h,v 1.19 1998/09/11 11:10:23 tomczyk Exp $
 */
class WaveFile : public AudioData {
//  friend class AFPlayer;
  public:
    explicit WaveFile(const string& filename);
    ~WaveFile() { }

    uint32 fileSize() const { return _fileSize; }
    uint32 type() const { return _type; }
    uint32 bitsInSample() const { return _bitsInSample; }

//    void playOn(AFPlayer& player, 
//                        const long length,
//                        const int gain = 0,
//                        const float delay = 0.1, 
//                        const int device = -1) const;

  protected:
    void readHeader();
    uint32 packBytes(Byte header[4]) const;

    bool isRiffHeader(Byte header[4]) const;
    bool isWaveHeader(Byte header[4]) const;
    bool isFmtHeader(Byte header[4]) const;
    bool isDataHeader(Byte header[4]) const;
    bool isOrdinaryFormatChunk(Byte header[4]) const;

    void readFormatTag(Byte header[2]);
    void readChannelCount(Byte header[2]);
    void readSamplingRate(Byte header[4]);
    void readBytesPerSecond(Byte header[4]);
    void readBytesPerSample(Byte header[2]);
    void readBitsPerSample(Byte header[2]);

    uint32 _fileSize;
    uint32 _type;
    uint32 _bitsInSample;

  private:
    WaveFile(const WaveFile&);
    WaveFile& operator=(const WaveFile&);
};

////////////////////////////////////////////////////////////

inline
bool WaveFile::isRiffHeader(Byte header[4]) const {
  return ((header[0] == 'R') && (header[1] == 'I')
       && (header[2] == 'F') && (header[3] == 'F'));
}

////////////////////////////////////////////////////////////

inline
bool WaveFile::isWaveHeader(Byte header[4]) const {
  return ((header[0] == 'W') && (header[1] == 'A')
       && (header[2] == 'V') && (header[3] == 'E'));
}

////////////////////////////////////////////////////////////

inline
bool WaveFile::isFmtHeader(Byte header[4]) const {
  return ((header[0] == 'f') && (header[1] == 'm')
       && (header[2] == 't') && (header[3] == ' '));
}

////////////////////////////////////////////////////////////

inline
bool WaveFile::isDataHeader(Byte header[4]) const {
  return ((header[0] == 'd') && (header[1] == 'a')
       && (header[2] == 't') && (header[3] == 'a'));
}

////////////////////////////////////////////////////////////

inline
bool WaveFile::isOrdinaryFormatChunk(Byte header[4]) const {
  return ((header[0] == 0x10) && (header[1] == 0x00)
       && (header[2] == 0x00) && (header[3] == 0x00));
}

////////////////////////////////////////////////////////////

inline
uint32 WaveFile::packBytes(Byte header[4]) const {
  return (header[0] + header[1] * 0x100
          + header[2] * 0x10000 + header[3] * 0x1000000);
}

////////////////////////////////////////////////////////////

//void WaveFile::playOn(AFPlayer& player, const long length,
//                const int gain, const float delay, 
//                const int device) const {
//  player.play(*this, length, gain, delay, device);
//}

////////////////////////////////////////////////////////////

inline
void WaveFile::readBytesPerSecond(Byte header[4]) {
  // do nothing, this information is not used
  return;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                     AudioFactory

/**
 * This class represents a kind of "factory" which creates all 
 * plattform dependend objects. So the user of aulib++ does not
 * need to take care of Alpha-Unix, SUN-Solaris, MS-DOS etc.<BR>
 * AudioFactory can also create AudioData objects for audio files.
 * The user does not need to know if they are SUN-au, WAV or somethig
 * else.
 *
 * @author <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>, <a href = "mailto:tomczyk@chili.wifo.uni-mannheim.de">Peter Tomczyk</a>
 * @version $Id: audata.h,v 1.19 1998/09/11 11:10:23 tomczyk Exp $
 */
class AudioFactory {

  public:

    /**
     * creates an AudioData object for the audio file having the given 
     * name. The user does not need to know if it is SUN-au, WAV or
     * something else.
     *
     * @param filename name of an audio file
     * @return         object representing the audio file
     *
     * @return Exception, if format is unknown or if reading faild.
     */
    static AudioData* openFile(string filename);

    //static AudioPlayer* createPlayer();

  private:

    /**
     * check if the file suffix is SUN-au suffix.
     *
     * @param filename name of a file
     * @return         true if the file has SUN-au suffix
     */
    static bool isAuFile(string filename);

    /**
     * check if the file suffix is WAV suffix.
     *
     * @param filename name of a file
     * @return         true if the file has WAV suffix
     */
    static bool isWavFile(string filename);
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       ByteReader

class ByteReader {
  public:
    virtual int bytesInSample() const = 0;
    virtual Sample toSample(const Byte bytes[]) const = 0;
};

////////////////////////////////////////////////////////////

class Lin8ByteReader : public ByteReader {
  public:
    int bytesInSample() const { return 1; }
    Sample toSample(const Byte bytes[]) const {
      return static_cast<char8>(bytes[0]); }
};

////////////////////////////////////////////////////////////

class MulawByteReader : public ByteReader {
  public:
    int bytesInSample() const { return 1; }
    Sample toSample(const Byte bytes[]) const {
      return ulaw2lin[static_cast<int>(bytes[0])]; }
  private:
    static const Sample ulaw2lin[];
};

////////////////////////////////////////////////////////////

class Lin16ByteReader : public ByteReader {
  public:
    int bytesInSample() const { return 2; }
    Sample toSample(const Byte bytes[]) const {
      return (static_cast<int>(bytes[0]) 
        + static_cast<int>(bytes[1]) * 0x100 - 32768); 
    }
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       ByteWriter

class ByteWriter {
  public:
    virtual int bytesInSample() const = 0;
    virtual void fromSample(const Sample x, Byte bytes[]) const = 0;
};

////////////////////////////////////////////////////////////

class Lin8ByteWriter : public ByteWriter {
  public:
    int bytesInSample() const { return 1; }
    void fromSample(const Sample x, Byte bytes[]) const {
      bytes[0] = static_cast<char8>(x); }
};

////////////////////////////////////////////////////////////

class MulawByteWriter : public ByteWriter {
  public:
    int bytesInSample() const { return 1; }
    void fromSample(const Sample x, Byte bytes[]) const {
     bytes[0] = lin2ulaw(x); }
  private:
    Byte lin2ulaw(Sample sample) const;
    static const int expLut[];
};

////////////////////////////////////////////////////////////

class Lin16ByteWriter : public ByteWriter {
  public:
    int bytesInSample() const { return 2; }
    void fromSample(const Sample x, Byte bytes[]) const {
      bytes[0] = static_cast<Byte>(x % 256);
      bytes[1] = static_cast<Byte>(x / 256);
    }
};

#endif

