/*
   This file is part of the Aulib library.
   Name: auops.cpp
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include "auops.h"
#include "error.h"
#include "aucache.h"
#include "audata.h"
#include "fft.h"
#include <cmath>
#include <algorithm>

using namespace std;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                      AuAmplitude

AuAmplitude::ValuePair AuAmplitude::calcCacheAmp(const AudioCache& cache, const Mode which) {

  long int  possum, posno, negsum, negno;
  Sample    maxpos, maxneg, minpos, minneg;
  ValuePair result(0, 0);

  switch (which) {
  /* if average must be calculated */
  case AVE:
    possum = negsum = 0;
    posno = negno = 0;

	for (size_t i = 0; i < cache.size(); i++) {
	  if (cache[i] > 0) {
	    possum = possum + cache[i];
	    posno++;
	  }
	  else {
	    if (cache[i] < 0) {
	      negsum = negsum + cache[i];
	      negno++;
		}
      }
	}

    if (posno==0) posno=1;
      result.first = possum/posno;
    if (negno==0) negno=1;
	  result.second = negsum/negno;

    break;

  /* if maximum must be calculated */
  case MAX:
    maxpos = maxneg = 0;

	for (size_t i = 0; i < cache.size(); i++)
	  if (cache[i] > maxpos)
	    maxpos = cache[i];
	  else
	    if (cache[i] < maxneg)
		  maxneg = cache[i];

	result.first = maxpos;
	result.second = maxneg;

    break;

  /* if minimum must be calculated */
  case MIN:
    minpos = MIN_AMP;
    minneg = MAX_AMP;

	for (size_t i = 0; i < cache.size(); i++)
	  if (cache[i] > 0) {
	    if (cache[i] < minpos)
	      minpos = cache[i];
	  }
	  else {
	    if (cache[i] < 0) {
	      if (cache[i] > minneg)
	        minneg = cache[i];
		}
	}

      if (minpos == MAX_AMP) minpos=0;
	  if (minneg == MIN_AMP) minneg=0;
      result.first = minpos;
      result.second = minneg;

	  break;

  /* if absolute must be calculated */
  case ABS:
	maxpos = maxneg = 0;

	for (size_t i = 0; i < cache.size(); i++)
	  if (cache[i] > maxpos)
		maxpos = cache[i];
	  else
	    if (cache[i] < maxneg)
		  maxneg = cache[i];

	result.first = 0;
	result.second = maxpos - maxneg;

    break;

  default:
		throw Exception("AuAmplitude::calcCacheAmp() - bad mode.");
  }

  /* return range */
  return result;

}

////////////////////////////////////////////////////////////

Sample AuAmplitude::maxAmp(const AudioCache& cache) {

  Sample max = 0;

  for (size_t i = 0; i < cache.size(); i++) {
    if (cache[i] > max)
      max = cache[i];
    else
      if (-cache[i] > max) max = -cache[i];
  }

  return max;
}

////////////////////////////////////////////////////////////

long int AuAmplitude::ampSum(const AudioCache& cache) {

  long int possum = 0;
  long int negsum = 0;

  for (size_t i = 0; i < cache.size(); i++)
    if (cache[i] > 0)
      possum = possum + cache[i];
    else
      if (cache[i] < 0)
        negsum = negsum + cache[i];

  return possum - negsum;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                      AuEnergy

double AuEnergy::energy(const AudioCache& cache) {
  double sum = 0.0;
  for (size_t i = 0; i < cache.size(); i++)
	  sum += 1.0 * cache[i] * cache[i] / cache.samplingRate();
  return sum;
}

////////////////////////////////////////////////////////////

double AuEnergy::power(const AudioCache& cache) {
  double sum = 0.0;
  for (size_t i = 0; i < cache.size(); i++)
	  sum += 1.0 * cache[i] * cache[i] / (2.0 * cache.size() + 1);
  return sum;
}

////////////////////////////////////////////////////////////

double AuEnergy::energyFFT(const FFTData& spec, const uint32 samplingRate) {
  const double divisor = 1.0 * samplingRate * samplingRate;
  double sum = 0.0;
  for (size_t i = 0; i < spec.size(); i++)
    sum += spec.amp(i) / divisor;
  return sum;
}

////////////////////////////////////////////////////////////

double AuEnergy::energyRange(const FFTData& spec, size_t from, size_t to) {
  double sum = 0.0;
  for (size_t i = from; i < to && i < spec.size(); i++)
    sum += spec.amp(i);
  return sum;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       AuFuF

const Float AuFuF::MAX_REL_AMP = 1.0;
const Float AuFuF::MAX_ABS_AMP = 1.0;
const size_t AuFuF::FIRST_SIGNIFICANT_FREQ = 10; // about 150Hz
const size_t AuFuF::FIRST_OVERTONE = 2;
const size_t AuFuF::LAST_OVERTONE = 6;
const size_t AuFuF::BASIC_NOISE_FREQ = 4;

////////////////////////////////////////////////////////////

FFTData AuFuF::filterNoise(const FFTData& spec, Float relativ, Float absolut) {

  Float maxval = 
    *max_element(spec.cbegin(&FFTData::amp), spec.cend(&FFTData::amp));
  double Relativ = ((maxval == 0.0) ? 0.0 : (0.01 * maxval * relativ));
  double Absolut = 0.01 * MAX_AMP * absolut;

  FFTData result(spec.size());

  for (size_t i = 0; i < spec.size(); i++) {
    if (spec.amp(i) < Relativ || spec.amp(i) < Absolut)
      result[i] = 0;
    else
      result[i] = spec[i];
  }

  return result;
}

////////////////////////////////////////////////////////////

void AuFuF::calcHarmonics(const FFTData& spec, FuFData* fufs, size_t freq1, size_t freq2, size_t freq) {

  size_t framesize = spec.size();
  size_t x = freq * freq2 / freq1;

  if (x < framesize - 1) {
    if ((*fufs)[x] > 0) {
      (*fufs)[1/freq1] += (*fufs)[freq] + (*fufs)[x];
      (*fufs)[x] = 0;
      (*fufs)[freq] = 0;
    }
    else {
      (*fufs)[freq] = spec.amp(freq);
    }
  }
}

////////////////////////////////////////////////////////////

AuFuF::FuFData AuFuF::calcFuf(const FFTData& spec) {

  size_t framesize = spec.size() / 2;
  FuFData fufs(framesize, 0);

  FFTData filtered = filterNoise(spec, MAX_REL_AMP, MAX_ABS_AMP);
  for (size_t i = FIRST_SIGNIFICANT_FREQ; i < framesize; i++) {
    if (filtered.amp(i) > 0) {
      for (size_t y = FIRST_OVERTONE; y < LAST_OVERTONE; y++)
        calcHarmonics(filtered, &fufs, y, y + 1, i);
    }
  }

  for (size_t i = 0; i < BASIC_NOISE_FREQ; i++)
    fufs[i] = 0;

  return fufs;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                      AuLength

size_t AuLength::length(const AudioCache& aucache, Format type, double fps) {

  const size_t size = aucache.size(); // in samples
  const uint32 bps = aucache.bytesInSample();
  const uint32 channels = aucache.channelCount();
  const uint32 sampRate = aucache.samplingRate();

  switch (type) {
    case BYTES:
      return samplesInBytes(size, bps, channels);
    case SECONDS:
      return static_cast<size_t>(ceil(samplesInSeconds(size, sampRate)));
    case FRAMES:
      return secondsInFrames(samplesInSeconds(size, sampRate), fps);
    default: // SAMPLES
      return size;
  }
}

////////////////////////////////////////////////////////////

size_t AuLength::length(const AudioData& audata, Format type, double fps) {

  const size_t size = audata.dataSize(); // in bytes
  const uint32 bps = audata.bytesInSample();
  const uint32 channels = audata.channelCount();
  const uint32 sampRate = audata.samplingRate();

  switch (type) {
    case SAMPLES:
      return bytesInSamples(size, bps, channels);
    case SECONDS:
      return static_cast<size_t>(ceil(samplesInSeconds(bytesInSamples(size, bps, channels), sampRate)));
    case FRAMES:
      return secondsInFrames(samplesInSeconds(bytesInSamples(size, bps, channels), sampRate), fps);
    default: // BYTES
      return size;
  }
}

////////////////////////////////////////////////////////////

double AuLength::convert(size_t val, Format from, Format to, uint32 bps, uint32 channels, uint32 sampRate, double fps) {

  switch (from) {

    case SAMPLES:
      switch (to) {
        case BYTES:
          return samplesInBytes(val, bps, channels);
        case SECONDS:
          return samplesInSeconds(val, sampRate);
        case FRAMES:
          return secondsInFrames(samplesInSeconds(val, sampRate), fps);
        default: // SAMPLES
          return val;
      }

    case SECONDS:
      switch (to) {
        case BYTES:
          return samplesInBytes(secondsInSamples(val, sampRate), bps, channels);
        case SAMPLES:
          return secondsInSamples(val, sampRate);
        case FRAMES:
          return secondsInFrames(val, fps);
        default: // SECONDS
          return val;
      }

    case FRAMES:
      switch (to) {
        case BYTES:
          return samplesInBytes(secondsInSamples(framesInSeconds(val, fps), sampRate), bps, channels);
        case SAMPLES:
          return secondsInSamples(framesInSeconds(val, fps), sampRate);
        case SECONDS:
          return framesInSeconds(val, fps);
        default: // FRAMES
          return val;
      }

    default: // BYTES
      switch (to) {
        case SAMPLES:
          return bytesInSamples(val, bps, channels);
        case SECONDS:
          return samplesInSeconds(bytesInSamples(val, bps, channels), sampRate);
        case FRAMES:
          return secondsInFrames(samplesInSeconds(bytesInSamples(val, bps, channels), sampRate), fps);
        default: // BYTES
          return val;
      }
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                      AuThreshold

int AuThreshold::sampsBelow(const AudioCache& cache, Sample threshold) {

  int numbelow = 0;

  for (size_t i = 0; i < cache.size(); i++)
	if (cache[i] < threshold)
      numbelow++;

  return numbelow;
}

////////////////////////////////////////////////////////////

int AuThreshold::sampsAbove(const AudioCache& cache, Sample threshold) {

  int numabove = 0;

  for (size_t i = 0; i < cache.size(); i++)
    if (cache[i] > threshold)
      numabove++;

  return numabove;
}
