/*
   This file is part of the Aulib library.
   Name: auops.h
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef _AU_OPERATORS_H_
#define _AU_OPERATORS_H_

#include "autypes.h"
#include "audata.h"
#include <pair.h>
#include <cstddef>
#include <cmath>
#include <vector>
using namespace std;

class AudioCache;
class FFTData;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                     AuAmplitude

/**
 * This class implements routines to get amplitude values.
 *
 * @author <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>, <a href = "mailto:tomczyk@chili.wifo.uni-mannheim.de">Peter Tomczyk</a>
 * @version $Id: auops.h,v 1.3 1998/10/09 07:24:19 tomczyk Exp tomczyk $
 */
class AuAmplitude {

  public:

    /**
     * Mode which can be used for <TT>calcCacheAmp</TT>.<BR>
     * MAX = maximal amplitude range<BR>
     * MIN = minimal amplitude range<BR>
     * AVE = average (above and below 0) range<BR>
     * ABS = absolute range (from 0 to sum of max pos and max neg amplitude)<BR>
     */
    enum Mode { MAX = 0, MIN = 1, AVE = 2, ABS = 3 };

    /**
     * Value Pair based on the template class pair (STL). It is used
     * by <TT>calcCacheAmp</TT> as return value.
     */
    typedef pair<Sample, Sample> ValuePair;

    /**
     * Function to calculate the range of the loaded
     * amplitudes in the audio cache and returns it depending on the
     * parameter <TT>which</TT>. <TT>which</TT> can be: MAX/MIN/AVE/ABS/SUM,
     * so that <TT>calcCacheAmp</TT> returns the maximal amplitude range, the minimal,
     * the average (above and below 0), an absolute range (from 0 to sum of
     * max pos and max neg amplitude) or the sum of all amplitudes (the
     * integral/the energy)
     *
     * @param cache audio cache containing the signal
     * @param which calculation mode
     * @return      the range (two values: positive first) as a <TT>ValuePair</TT>
     *
     * @exception Exception, if the mode is unknown
     */
    static ValuePair calcCacheAmp(const AudioCache& cache, const Mode which);

    /**
     * Returns maximal amplitude range.
     */
    static Sample maxAmp(const AudioCache& cache);

    /**
     * Returns the sum of all amplitudes (the integral/the energy).
     */
    static long int ampSum(const AudioCache& cache);
};


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                     AuEnergy

/**
  * This class implements routines to calculate signal energy and power 
  * in time and frequency domain.
  *
  * @author <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>, <a href = "mailto:tomczyk@chili.wifo.uni-mannheim.de">Peter Tomczyk</a>
  * @version $Id: auops.h,v 1.3 1998/10/09 07:24:19 tomczyk Exp tomczyk $
  */
class AuEnergy {

  public:

    /**
      * Function to calculate signal energy on the samples stored in the
      * audio cache. The original definition of the signal energy of a
      * discrete-time signal is:<BR>
      * E = lim{n->infty} 1/sampling_rate * sum{-n to n}( sqr( abs(x(n)) ) )
      *
      * @param cache audio cache containing the signal
      * @return      calculated energy
      */
    static double energy(const AudioCache& cache);

    /**
      * Function to calculate signal power on the samples stored in the
      * audio cache. The original definition of the signal power of a
      * discrete-time signal is:<BR>
      * P = lim{n->infty} 1/(2*n+1) * sum{-n to n}( sqr( abs(x(n)) ) )
      * We have restricted n to the size of the audio cache!
      *
      * @param cache audio cache containing the signal
      * @return      calculated power
      */
    static double power(const AudioCache& cache);

    /**
      * Function to calculate signal energy on the complex samples 
      * stored in the wave vector. The original definition is:<BR>
      * E = Integral{-samplingRate/2 to samplingRate/2}
      *             ( sqr(1/samplingRate) * sqr(abs(X(f))) ) df
      *
      * @param wave         vector of complex samples (= signal)
      * @param samplingRate used sampling rate
      * @return             calculated energy
      */
    static double energyFFT(const FFTData& spec, const uint32 samplingRate);

    /**
     * Function to calculate signal energy on the samples stored 
     * in the audio cache, restricting only to frequency bands 
     * between from and to.
     */
    static double energyRange(const FFTData& spec, size_t from, size_t to);

};


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                         AuFuF

class AuFuF {
  public:
    typedef vector<Float> FuFData;

    /**
     * filters away low amplitude components of the FFT spectrum,
     * which constitue noise. Prepares FFT spectrum for fuf calculation.
     *
     * param@ spec    reference to an array of frequency values
     * param@ realtiv amplitude values are zeroed, if they are less than
     *                "relativ %" under the max amp value in the spectrum
     * param@ absolut amplitude values are zeroed, if they are less than
     *                "absolut %" under the max possible amp value
     */
    static FFTData filterNoise(const FFTData& spec, Float relativ, Float absolut);

    /**
     * calculates harmonics of a given frequency spectrum
     * (detect repetative frequencies in interval distances).
     * Required for fuf calculation.
     * FREQ1 and FREQ2 signify the interval (represented by a lower frequency
     * FREQ and a higher frequency x) of the overtones to be tested, i.e.
     * <P><PRE>
     * 2 & 3 signify a Fifth  => fuf = Octave below lower frequency
     * 3 & 4 signify a Fourth => fuf = Duodecime below lower freq
     * 4 & 5 signify a Maj. Third => fuf = 2 Octaves below lower freq
     * 5 & 6 signify a Min. Third => fuf = 2 Decimes below lower freq
     * </PRE>
     * param@ spec  reference to an array of frequency values
     * param@ fufs  reference to an array with collection of fufs
     * param@ freq1 signify the interval of the overtones to be tested
     * param@ freq2 signify the interval of the overtones to be tested
     * param@ freq  the band of the lower frequency to be tested for overtones
     */
    static void calcHarmonics(const FFTData& spec, FuFData* fufs, size_t freq1, size_t freq2, size_t freq);
    /**
     * calculates an array of fundamental frequencies for a FFT Spectrum
     * Results in an  array of size win_size/2 with fufs.
     *
     * @param spec reference to an array of frequency values
     * @return fuf array
     */
    static FuFData calcFuf(const FFTData& spec);
  private:

    ///
    static const Float MAX_REL_AMP;

    ///
    static const Float MAX_ABS_AMP;

    ///
    static const size_t FIRST_SIGNIFICANT_FREQ;

    ///
    static const size_t FIRST_OVERTONE;

    ///
    static const size_t LAST_OVERTONE;

    /// highest frequency of basic noice.
    static const size_t BASIC_NOISE_FREQ;
};


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       AuLength

class AudioCache;
class AudioData;

/**
  * This class represents an audio tool which figures out length of an audio file 
  * or an audio cache.
  *
  * @author <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>, <a href = "mailto:tomczyk@chili.wifo.uni-mannheim.de">Peter Tomczyk</a>
  * @version $Id: auops.h,v 1.3 1998/10/09 07:24:19 tomczyk Exp tomczyk $
  */
class AuLength {
  public:
    enum Format { BYTES, SAMPLES, SECONDS, FRAMES };

    static size_t length(const AudioCache& aucache, Format type, double fps = FPS_C);
    static size_t length(const AudioData& aucache, Format type, double fps = FPS_C);

    static size_t bytesInSamples(size_t bytes, uint32 bps, uint32 channels);
    static size_t samplesInBytes(size_t samps, uint32 bps, uint32 channels);
    static size_t secondsInSamples(double secs, uint32 sampRate);
    static double samplesInSeconds(size_t samps, uint32 sampRate);
    static size_t secondsInFrames(double secs, double fps = FPS_C);
    static double framesInSeconds(size_t frames, double fps = FPS_C);

    static double convert(size_t val, Format from, Format to, uint32 bps, uint32 channels, uint32 sampRate, double fps = FPS_C);
};

////////////////////////////////////////////////////////////

inline
size_t AuLength::bytesInSamples(size_t bytes, uint32 bps, uint32 channels) {
  return (bytes / (bps * channels));
}

inline
size_t AuLength::samplesInBytes(size_t samps, uint32 bps, uint32 channels) {
  return (samps * bps * channels);
}

inline
size_t AuLength::secondsInSamples(double secs, uint32 sampRate) {
  return static_cast<size_t>(ceil(secs * sampRate));
}

inline
double AuLength::samplesInSeconds(size_t samps, uint32 sampRate) {
  return (1.0 * samps / sampRate);
}

inline
size_t AuLength::secondsInFrames(double secs, double fps) {
  return static_cast<size_t>(ceil(secs * fps));
}

inline
double AuLength::framesInSeconds(size_t frames, double fps) {
  return (frames / fps);
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                     AuThreshold

/**
  * This class implements routines to calculate amplitude thresholds on audio cache.
  *
  * @author <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>, <a href = "mailto:tomczyk@chili.wifo.uni-mannheim.de">Peter Tomczyk</a>
  * @version $Id: auops.h,v 1.3 1998/10/09 07:24:19 tomczyk Exp tomczyk $
  */
class AuThreshold {

  public:

    /**
      * Function to calculate how many of the loaded amplitudes in the 
      * audio cache are below a given threshold.
      *
      * @param cache     audio cache
      * @param threshold threshold to be used
      * @return          number of samples
      */
    static int sampsBelow(const AudioCache& cache, Sample threshold);

    /**
      * Function to calculate how many of the loaded amplitudes in the 
      * audio cache are above a given threshold.
      *
      * @param cache     audio cache
      * @param threshold threshold to be used
      * @return          number of samples
      */
    static int sampsAbove(const AudioCache& cache, Sample threshold);
};

#endif
