/*
   This file is part of the Aulib library.
   Name: austruct.h
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef _AUSTRUCT_H_
#define _AUSTRUCT_H_

#include "error.h"
#include <cstddef>

#ifdef OLDGNU
#include <iostream.h>
#else
#include <iostream>
#endif

using namespace std;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       Segment

/// represents positions.
typedef unsigned long int pos_t;

/**
 * Segment encapsulates two positions, the start and the
 * end of a segment.
 *
 * @author Peter Tomczyk
 * @version $Id: austruct.h,v 1.1 1998/09/11 11:15:07 tomczyk Exp $
 */
class Segment {
public:
  explicit Segment(pos_t b = 0, pos_t e = 0) : begin(b), end(e) { }
  pos_t begin;
  pos_t end;
};

ostream& operator<<(ostream& os, const Segment& s);

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                          Array

/**
 * The Array class supports a simply array. Arrays can
 * be resized but 'resize' works destructive, so they 
 * are not realy dynamic containers. Array elements
 * must support the following interface: default constructor,
 * copy constructor, destructor, operator=, operator<<.
 *
 * @author Silvia Pfeiffer, Peter Tomczyk
 * @version $Id: austruct.h,v 1.1 1998/09/11 11:15:07 tomczyk Exp $
 */
template<class T>
class Array {
  friend ostream& operator<<(ostream& os, const Array& a) {
    return a.printOn(os);
  }

  public:

    /// Type of the value stored in the container.
    typedef T value_type;

    /// Reference to an element.
    typedef T& reference;

    /// Const reference to an element.
    typedef const T& const_reference;

    /// Type representing size.
    typedef size_t size_type;
    
    /**
     * Creates an Array that has 'size' elements.
     *
     * @param size Array size
     * @exception Exception
     */
    explicit Array(size_type size = 0);

    /**
     * Creates an Array that contains a copy of the
     * specified Array.
     *
     * @param a another Array
     * @exception Exception
     */
    Array(const Array& a) : _size(a._size), _data(0) { copy(a); }

    /// Destroys the Array.
    virtual ~Array() { destroy(); }

    Array& operator=(const Array& a);

    /**
     * Returns true if the elements of the invoking Array
     * are all equal to the elements of 'a'.
     */
    virtual bool operator==(const Array& a) const;

    /// Returns !(operator==(a)). Should not be overridden.
    bool operator!=(const Array& a) const;

    /// Returns an const reference to the element specified by i.
    const T& operator[](size_type i) const { return _data[i]; }

    /// Returns an reference to the element specified by i.
    T& operator[](size_type i) { return _data[i]; }

    /**
     * Changes the size of the Array to that specified by 'size'.
     * Works destructive.
     *
     * @param size new Array size
     * @exception Exception
     */
    void resize(size_type size);

    /// Returns the number of elements current in the Array.
    size_type size() const { return _size; }

    /// Returns true if the invoking Array is empty.
    bool empty() const { return _size == 0; }

    class Iter {
      //friend class ConstIter;
      public:
        Iter(T* t = 0) : _cur(t) { }
        Iter& operator=(const Iter& iter) { 
          _cur = iter._cur; return *this; }
        bool operator==(const Iter& iter) const {
          return _cur == iter._cur; }
        bool operator!=(const Iter& iter) const {
          return _cur != iter._cur; }
        T& operator*() { return *_cur; }
        T* operator->() { return _cur; }
        Iter& operator++() { _cur++; return *this; }
        Iter operator++(int) { Iter tmp = *this; _cur++; return tmp; }
        Iter& operator--() { _cur--; return *this; }
        Iter operator--(int) { Iter tmp = *this; _cur--; return tmp; }
      private:
        T* _cur;
    };

    class ConstIter {
      public:
        ConstIter(T* t = 0) : _cur(t) { }
        //ConstIter(const Iter& iter) : _cur(iter._cur) { }
        ConstIter& operator=(const ConstIter& iter) { 
          _cur = iter._cur; return *this; }
        bool operator==(const ConstIter& iter) const {
          return _cur == iter._cur; }
        bool operator!=(const ConstIter& iter) const {
          return _cur != iter._cur; }
        const T& operator*() { return *_cur; }
        const T* operator->() { return _cur; }
        ConstIter& operator++() { _cur++; return *this; }
        ConstIter operator++(int) { 
          ConstIter tmp = *this; _cur++; return tmp; }
        ConstIter& operator--() { _cur--; return *this; }
        ConstIter operator--(int) { 
          ConstIter tmp = *this; _cur--; return tmp; }
      private:
        T* _cur;
    };

    /// Iterator.
    typedef Iter iterator;

    /// Const iterator.
    typedef ConstIter const_iterator;

    /// Returns an iterator to the first element of the Array.
    Iter begin() { return _data; }

    /// Returns an iterator to the last element of the Array.
    Iter end() { return _data + _size; }

    /// Returns an const iterator to the first element of the Array.
    ConstIter begin() const { return _data; }

    /// Returns an const iterator to the last element of the Array.
    ConstIter end() const { return _data + _size; }

  protected:

    /// Assigns the Array the contents of 'a'.
    void copy(const Array& a);

    /// Destroys the Array data.
    void destroy() { delete[] _data; }

    /// Prints an info about the Array on the given stream.
    virtual ostream& printOn(ostream& os) const;

    /// Number of elements current in the Array.
    size_type _size;

    /// Array data.
    T* _data;
};

////////////////////////////////////////////////////////////

template<class T>
Array<T>::Array(size_t size) : _size(size), _data(0) {
  if (size != 0) {
    _data = new T[size];
    if (_data == 0)
      throw Exception("Array::Array() - cannot allocate memory.");
  }
}

////////////////////////////////////////////////////////////

template<class T>
Array<T>& Array<T>::operator=(const Array<T>& array) {
  if (this != &array) {
    destroy();
    copy(array);
  }
  return *this;
}

////////////////////////////////////////////////////////////

template<class T>
bool Array<T>::operator==(const Array<T>& a) const {
  if (_size != a._size)
    return false;
  for (size_type i = 0; i < _size; ++i) {
    if (_data[i] != a._data[i])
      return false;
  }
  return true;
}

////////////////////////////////////////////////////////////

template<class T>
bool Array<T>::operator!=(const Array<T>& a) const {
  return !(operator==(a));
}

////////////////////////////////////////////////////////////

template<class T>
void Array<T>::copy(const Array<T>& a) {
  if ((_size = a._size) != 0) {
    _data = new T[_size];
    if (_data == 0)
      throw Exception("Array::Array() - cannot allocate memory.");
    for (size_type i = 0; i < _size; ++i)
      _data[i] = a._data[i];
  }
}

////////////////////////////////////////////////////////////

template<class T>
void Array<T>::resize(size_t size) {
  destroy();
  if ((_size = size) != 0) {
    _data = new T[_size];
    if (_data == 0)
      throw Exception("Array::resize() - cannot allocate memory.");
  }
}

////////////////////////////////////////////////////////////

template<class T>
ostream& Array<T>::printOn(ostream& os) const {
  os << "[ Array, size = " << _size << ", data: ";
  for (size_type i = 0; i < _size; i++)
    cout << _data[i] << " ";
  cout << "]";
  return os;
}

#endif

