/*
   This file is part of the Aulib library.
   Name: bg.h
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef _AU_BACKGROUND_H_
#define _AU_BACKGROUND_H_

#include "autypes.h"
#include "austruct.h"
#include "error.h"
#include "auops.h"
#include "aucache.h"
#include "audata.h"
#include "fft.h"
#include <cstddef>
#include <vector>
#include <algorithm>

#ifdef OLDGNU
#include <iostream.h>
#else
#include <iostream>
#endif

using namespace std;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                  BGFinders


/**
 * BGFinder looks for quiet segments inside of audio
 * files. It splits an audio file into a number of windows
 * and estimates the loudness of each window.
 * If the calculated window loudness is less than a limit,
 * a quiet window was detected. Overlapping quiet windows
 * are put together, they form quiet segments.
 * <p>
 * The window size is defined by 'winSize'. 'winStep' 
 * defines the distance between the start of a window and 
 * the start of the next one. 'relLimit' defines the 
 * loudness limit: limit := relLimit * max loudness { all
 * windows }
 *
 * @author Peter Tomczyk
 * @version $Id: bg.h,v 1.5 1998/09/11 11:17:24 tomczyk Exp $
 */
class BGFinder {

  friend ostream& operator<<(ostream& os, const BGFinder& bf) {
    return bf.printOn(os);
  }

  public:

    /// represents loudness
    typedef double Loudness;

    /**
     * Creates new BGFinder.
     *
     * @param size window size
     * @param step distance between windows
     * @param relLimit maximum relative loudness
     */
    BGFinder(size_t size = 64, size_t step = 64, Float relLimit = 0.5);

    /// Destroys BGFinder.
    virtual ~BGFinder() { }

    /// Returns current window size.
    size_t winSize() const { return _winSize; }

    /// Returns distance between windows.
    size_t winStep() const { return _winStep; }

    /// Returns maximum relative loudness.
    Float relLimit() const { return _relLimit; }

    /// Sets window size.
    void winSize(size_t size) { _winSize = size; }

    /// Sets distance between windows (0 < step <= winSize).
    void winStep(size_t step) { _winStep = step; }

    /// Sets the relative loudness limit (between 0 and 1).
    void relLimit(Float limit) { _relLimit = limit; }

    /**
     * Looks for quiet segments in the given audio file. 
     * Returns a vector of segments.
     *
     * @param audata audio file
     * @return vector of segments (may be empty)
     */
    vector<Segment> findQuietSegments(const AudioData& audata) const;

  protected:

    /**
     * Calculates the maximum loudness a window may have
     * to be a quiet window.
     *
     * @param louds vector containing windows loudness
     * @return maximum absolute loudness or 0 if
     *         the given vector is empty
     */
    Loudness absLimit(const vector<Loudness>& louds) const;

    /**
     * Estimates the loudness of each window. Returns
     * vector of calculated values.
     *
     * @param audata audio file should be analysed
     * @return vector of values: loudness[window_number]
     */
    vector<Loudness> calcValues(const AudioData& audata) const;

    /** 
     * Calculates the bounds of quiet segments. Each window
     * having loudness < limit makes a quiet segment.
     * Overlapping segments are put together.
     *
     * @param values vector containing the loudness of each window
     * @return vector containing quiet segments
     */
    vector<Segment> calcPositions(const vector<Loudness>& values) const;

    /// Estimates the loudness of the window stored in 'cache'.
    virtual Loudness opLoudness(const AudioCache& cache) const = 0;

    /// Writes an info about the finder on the given stream.
    virtual ostream& printOn(ostream& os) const;

    /// Window size.
    size_t _winSize;

    /// Distance between windows.
    size_t _winStep;

    /// Maximum relative loudness of a window to be a quiet window [%].
    Float _relLimit;
};

inline
vector<Segment> BGFinder::findQuietSegments(const AudioData& audata) const {
  return calcPositions(calcValues(audata));
}

inline
BGFinder::Loudness BGFinder::absLimit(const vector<Loudness>& louds) const {
  vector<Loudness>::const_iterator itr
    = max_element(louds.begin(), louds.end());
  return (itr == louds.end() ? 0 : *itr * _relLimit);
}

/**
 * Concrete BGFinder using maximum window amplitude
 * to estimate the window loudness.
 *
 * @author Peter Tomczyk
 * @version $Id: bg.h,v 1.5 1998/09/11 11:17:24 tomczyk Exp $
 */
class BGFinderAMP : public BGFinder {
  public:
    BGFinderAMP(size_t size, size_t step, Float cutoff)
      : BGFinder(size, step, cutoff) { }
 protected:
   /// Returns maximum amplitude in 'cache'.
   Loudness opLoudness(const AudioCache& cache) const {
     return 1.0L * AuAmplitude::maxAmp(cache);
   }
};

/**
 * Concrete BGFinder using window energy
 * to estimate the window loudness.
 *
 * @author Peter Tomczyk
 * @version $Id: bg.h,v 1.5 1998/09/11 11:17:24 tomczyk Exp $
 */
class BGFinderENG : public BGFinder {
  public:
    BGFinderENG(size_t size, size_t step, Float cutoff)
      : BGFinder(size, step, cutoff) { }
 protected:
   /// Returns energy of the window stored in 'cache'.
   Loudness opLoudness(const AudioCache& cache) const {
     return AuEnergy::energy(cache);
   }
};

/**
 * Concrete BGFinder using window power
 * to estimate the window loudness.
 *
 * @author Peter Tomczyk
 * @version $Id: bg.h,v 1.5 1998/09/11 11:17:24 tomczyk Exp $
 */
class BGFinderPOW : public BGFinder {
  public:
    BGFinderPOW(size_t size, size_t step, Float cutoff)
      : BGFinder(size, step, cutoff) { }
 protected:
   /// Returns power of the window stored in 'cache'.
   Loudness opLoudness(const AudioCache& cache) const {
     return AuEnergy::power(cache);
   }
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                 BGAnalyser


/**
 * BGAnalyser makes statistics about the frequency
 * spectrum of an audio file. It takes an audio file
 * and a number of segments, which specify positions
 * inside of the file. Then it calculates the
 * frequency spectrum at these positions and sums
 * up the results.
 *
 * @author Peter Tomczyk
 * @version $Id: bg.h,v 1.5 1998/09/11 11:17:24 tomczyk Exp $
 */
class BGAnalyser {

  friend ostream& operator<<(ostream& os, const BGAnalyser& ba) {
    return ba.printOn(os);
  }

  public:

    /// Creates default BGAnalyser.
    BGAnalyser();

    /**
     * Creates new BGAnalyser.
     *
     * @param step    distance between two FFT windows
     * @param fftTool tool performing FFT
     * @param func    complex-to-real convertion method
     * @exception Exception
     */
    BGAnalyser(size_t step, FFT* fftTool, FFTData::ConvFunc func);

    /// Destroys BGAnalyser.
    virtual ~BGAnalyser();

    /// Returns distance between FFT windows.
    size_t winStep() const { return _winStep; }

    /// Returns const reference to the FFT tool.
    const FFT& fftTool() const { return *_fftTool; }

    /// Returns complex-to-real conversion method.
    FFTData::ConvFunc convFunc() const { return _convFunc; }

    /// Sets distance between FFT windows (0 < step <= winSize)
    void winStep(size_t step) { _winStep = step; }

    /// Returns reference to the FFT tool.
    FFT& fftTool() { return *_fftTool; }

    /// Sets complex-to-real conversion method.
    void convFunc(FFTData::ConvFunc f) { _convFunc = f; }

    /**
     * Returns statistics about the frequency spectrum of
     * the given audio file. 'segments' specifies the 
     * segments of 'audata' which should be analysed.
     * Example:
     * <p>
     * There is an AudioData object 'audio1' containing
     * 1000 samples. 'segments' specifies the following
     * two segments: [sample no. 64 - sample no. 128)
     * and [sample no. 256 - sample no. 512). FFT uses
     * window size of 64 samples and window step is also 
     * 64, so FFT windows do not overlap.
     * <p>
     * Inside of the first segment only one frequency
     * spectrum can be calculated, because this segment
     * contains exactly 64 samples. Inside of the second
     * segment 4 spectrums will be calculated (size = 256).
     * 'calcSpectrum()' returns the sum of the 5 frequency
     * spectrums as result.
     *
     * @param audata   audio file
     * @param segments a number of segments
     * @return the sum of all calculated frequency spectrums
     */
    FFTData analyse(const AudioData& audata,
      const vector<Segment>& segments) const;

  protected:

    /**
     * To avoid overflow errors each frequency spectrum
     * must be normalized before it is added to the
     * other ones. This method calculates the multiplier.
     * multiplier := 1.0 / number of FFTs will be done
     *
     * @param segments vector of segments
     * @return multiplier
     */
    Float calcNorm(const vector<Segment>& segments) const;

    /**
     * Adds the frequency spectrum of the window stored in
     * 'cache' to the frequency spectrum stored in 'spec'.
     * Multiplies each frequency to norm to avoid overflow.
     *
     * @param cache AudioCache containing a window
     * @param spec frequency spectrum to be updated
     * @param norm multipier to avoid overflow error
     */
    void updateSpec(const AudioCache& cache, FFTData* spec, Float norm) const;

    /// Prints an info about the analyser on the given stream.
    virtual ostream& printOn(ostream& os) const;

    /// Distance between start of a FFT window and start of the next one.
    size_t _winStep;

    /// Tool performing FFT.
    FFT* _fftTool;

    /// Temporary frequency spectrum used for better performance.
    mutable FFTData _tempSpec;

    /// complex-to-real convertion method.
    FFTData::ConvFunc _convFunc;

  private:

    BGAnalyser(const BGAnalyser&);
    BGAnalyser& operator=(const BGAnalyser&);
};

////////////////////////////////////////////////////////////

inline
BGAnalyser::BGAnalyser() : _winStep(1), _fftTool(new FFT()), 
  _tempSpec(FFTData(0)), _convFunc(&FFTData::amp) {

  if (_fftTool == 0)
    throw Exception("BGAnalyser::BGAnalyser() - cannot allocate memory.");
  _tempSpec.resize(_fftTool->winSize());
}

////////////////////////////////////////////////////////////

inline
BGAnalyser::BGAnalyser(size_t step, FFT* fftTool, FFTData::ConvFunc f)
  : _winStep(step), _fftTool(fftTool), _tempSpec(FFTData(0)), _convFunc(f) {

  if (_fftTool == 0)
    throw Exception("BGAnalyser::BGAnalyser() - FFT tool is missing.");
  _tempSpec.resize(_fftTool->winSize());
}

////////////////////////////////////////////////////////////

inline
BGAnalyser::~BGAnalyser() { delete _fftTool; }

////////////////////////////////////////////////////////////

inline
void BGAnalyser::updateSpec(const AudioCache& cache, FFTData* spec, Float norm) const {
  _fftTool->fft(cache, &_tempSpec);
  for (size_t i = 0; i < _fftTool->winSize(); i++)
    (*spec)[i] += norm * (_tempSpec.*_convFunc)(i);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                    BGExtractor

/**
 * BGExtractor removes frequencies form the spectrum of a
 * signal. The user uses a vector<bool> to specify which
 * frequencies should be removed.
 * vector<bool>[i] = true means "do not remove",
 * vector<bool>[i] = false means "remove".
 *
 * @author Peter Tomczyk
 * @version $Id: bg.h,v 1.5 1998/09/11 11:17:24 tomczyk Exp $
 */
class BGExtractor {

  public:

    /// Creates default BGExtractor.
    BGExtractor();

    /**
     * Creates new BGExtractor.
     *
     * @param freqs specifies frequencies to remove
     * @param func  windowing function
     */
    BGExtractor(vector<bool> freqs, FFT::WinFunc func = FFT::NONE);

    /// Destroys BGExtractor.
    virtual ~BGExtractor();

    /// Returns vector of specified frequencies.
    const vector<bool>& freqsToRemove() const { return _freqs; }

    /// Returns windowing function.
    FFT::WinFunc winFunc() const { return _fftTool->winFunc(); }

    /// Set frequencies to remove.
    void freqsToRemove(const vector<bool>& freqs);

    /// Sets windowing function.
    void winFunc(FFT::WinFunc func);

    /**
     * Removes all specified frequencies from in-cache
     * and stores the filtered signal into the out-cache.
     * In and out don't need to be different. Cache size
     * should be close to the size of _freqs.
     *
     * @param in input signal
     * @param out signal containing only background frequencies
     */
    void extract(const AudioCache& in, AudioCache* out) const;

  protected:

    void copyIntoCache(const FFTData& wave, AudioCache* cache) const;
    /// Turns the floating point value 'f' to an integer Sample value.
    Sample round(Float f) const;

    /// Prints an info about the extractor on the given stream.
    virtual ostream& printOn(ostream& os) const;

    /// Specifies frequencies to remove.
    vector<bool> _freqs;

    /// Tool performing FFT.
    FFT* _fftTool;

    /// Temprary spectrum used for better performance.
    mutable FFTData _tempSpec;

  private:
    BGExtractor(const BGExtractor&);
    BGExtractor& operator=(const BGExtractor&);
};

////////////////////////////////////////////////////////////

inline
BGExtractor::BGExtractor(vector<bool> freqs, FFT::WinFunc func = FFT::NONE)
  : _freqs(freqs), _fftTool(new FFT(freqs.size(), func)),
  _tempSpec(FFTData(0)) {

  if (_fftTool == 0)
    throw Exception("BGExtractor::BGExtractor() - cannot allocate memory.");
  _tempSpec.resize(_fftTool->winSize());
}

////////////////////////////////////////////////////////////

inline
BGExtractor::~BGExtractor() { delete _fftTool; }

////////////////////////////////////////////////////////////

inline
void BGExtractor::freqsToRemove(const vector<bool>& freqs) {
  _freqs = freqs;
  _fftTool->set(_freqs.size(), _fftTool->winFunc());
  _tempSpec.resize(_fftTool->winSize());
}

////////////////////////////////////////////////////////////

inline
void BGExtractor::winFunc(FFT::WinFunc func) {
  _fftTool->set(_fftTool->winSize(), func); }

//////////////////////////////////////////////////////////////

inline
void BGExtractor::copyIntoCache(const FFTData& wave, AudioCache* cache) const {
  for (size_t i = 0; i < wave.size() && i < cache->size(); i++)
    (*cache)[i] = round(wave.amp(i));
}

////////////////////////////////////////////////////////////

inline
Sample BGExtractor::round(Float f) const { 
  return static_cast<Sample>(floor(f + 0.5)); 
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                      BGUtil

/**
 * BGUtil implements some simply utilities needed by
 * background extraction.
 *
 * @author Peter Tomczyk
 * @version $Id: bg.h,v 1.5 1998/09/11 11:17:24 tomczyk Exp $
 */
class BGUtil {
  public:

    /**
     * Calculates which frequencies of 'spec' are background
     * frequencies and which not. A background frequency must
     * be louder than 'relLimit' percent of the loudest frequency.
     *
     * @param spec frequency spectrum
     * @param func complex-to-real conversion method
     * @param relLimit minimum relative loudness (0 < reLimit < 1)
     */
    static vector<bool> bgFreqs(const FFTData& spec, FFTData::ConvFunc func, Float relLimit);
};

#endif

