/*
   This file is part of the Aulib library.
   Name: fft.h
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#ifndef _FFT_H_
#define _FFT_H_

//
// This is not original by me, but adopted from a packet by
// Steve Haehnichen. I also include his comments therefore.
//
// <a href = "mailto:pfeiffer@pi4.informatik.uni-mannheim.de">Silvia Pfeiffer</a>

/*
 * fft.c: Code for generic fft routines.
 * [Part of simple-fft-1.5.tar.Z (ver 1.5)]
 *
 * Copyright (C) 1991 Steve Haehnichen.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  fft.c was derived from US Navy radar code written by Steve Sampson
 *  in June, 1988 for the MS-DOS platform, and placed in the Public Domain.
 *  The fft() function was copied directly from fft.c in his FFT26.ZIP.
 *    Steve Sampson
 *    Box 45668
 *    Tinker AFB, OK, 73145
 *
 *  Steve Haehnichen  (shaehnic@ucsd.edu)      March, 1992.
 *
 */

#include "autypes.h"
#include "austruct.h"
#include "error.h"
#include "aucache.h"
#include <complex>
#include <cmath>
#include <cstddef>

#ifdef OLDGNU
#include <iostream.h>
#else
#include <iostream>
#endif

using namespace std;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       FFTUtil

/**
 * The FFTUtil class implements some simply functions used
 * by FFT tools.
 *
 * @author Silvia Pfeiffer, Peter Tomczyk
 * @version $Id: fft.h,v 1.1 1998/09/11 11:19:21 tomczyk Exp tomczyk $
 */
class FFTUtil {
public:

  /// Returns bits = Max { b | 2^b <= samples }.
  static size_t samplesInBits(size_t samples);

  /// Returns 2^bits, if bits == 0 returns 0.
  static size_t bitsInSamples(size_t bits) { 
    return ((bits == 0) ? 0 : (1UL << bits)); }

  /// freqIndex = min { i : i * sampRate / nfreqs >= freq }
  static size_t freqIndex(size_t freq, size_t nfreqs, uint32 sampRate);
};

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       FFTData

class FFTData : public Array<complex<Float> > {
  friend class BaseFFT;
  friend FFTData operator+(const FFTData&, const FFTData&);
  friend FFTData operator-(const FFTData&, const FFTData&);
  friend FFTData operator*(Float, const FFTData&);
  public:
    typedef Float base_type;
    typedef Float (FFTData::*ConvFunc)(size_type) const;

    explicit FFTData(size_type size = 0) 
      : Array<complex<Float> >(size) { } 

    FFTData& operator=(const FFTData& d) {
      Array<complex<Float> >::operator=(d);
      return *this;
    }

    Float real(size_type i) const;
    Float sqr(size_type i) const;
    Float amp(size_type i) const;
    Float dB(size_type i) const;
    Float phase(size_type i) const;
    Float add(size_type i) const;

    class CIter {
    public:
      CIter(const FFTData& cc, FFTData::ConvFunc f, 
        FFTData::size_type i) : c(&cc), op(f), pos(i) { }
      CIter& operator=(const CIter& iter);
      bool operator==(const CIter& iter) const;
      bool operator!=(const CIter& iter) const;
      Float operator*() const;
      CIter& operator++();
      CIter operator++(int);
      CIter& operator--();
      CIter operator--(int);
    private:
      const FFTData* c;
      FFTData::ConvFunc op;
      FFTData::size_type pos;
    };

    CIter cbegin(ConvFunc op) const { 
      return CIter(*this, op, 0); }
    CIter cend(ConvFunc op) const { 
      return CIter(*this, op, _size); }
};

////////////////////////////////////////////////////////////

inline
Float FFTData::real(size_type i) const { 
  return _data[i].real(); }

////////////////////////////////////////////////////////////

inline
Float FFTData::sqr(size_type i) const { 
  return norm(_data[i]); }

////////////////////////////////////////////////////////////

inline
Float FFTData::amp(size_type i) const { 
  return abs(_data[i]); }

////////////////////////////////////////////////////////////

inline
Float FFTData::dB(size_type i) const { 
  return ((amp(i) > 0) ? 10*log10(amp(i)) : 10*log10(0.000001)); }

////////////////////////////////////////////////////////////

inline
Float FFTData::phase(size_type i) const { 
  return arg(_data[i]); }

////////////////////////////////////////////////////////////

inline
Float FFTData::add(size_type i) const { 
  return _data[i].real() + _data[i].imag(); }

////////////////////////////////////////////////////////////

inline
FFTData::CIter& FFTData::CIter::operator=(const CIter& iter) {
  c = iter.c; op = iter.op; pos = iter.pos; return *this; }

////////////////////////////////////////////////////////////

inline
bool FFTData::CIter::operator==(const CIter& iter) const {
  return (c == iter.c && op == iter.op && pos == iter.pos); }

////////////////////////////////////////////////////////////

inline
bool FFTData::CIter::operator!=(const CIter& iter) const {
  return (!operator==(iter)); }

////////////////////////////////////////////////////////////

inline
Float FFTData::CIter::operator*() const { 
  return (c->*op)(pos); }

////////////////////////////////////////////////////////////

inline
FFTData::CIter& FFTData::CIter::operator++() {
  pos++; return *this; }

////////////////////////////////////////////////////////////

inline
FFTData::CIter FFTData::CIter::operator++(int) {
  CIter tmp = *this; pos++; return tmp; }

////////////////////////////////////////////////////////////

inline
FFTData::CIter& FFTData::CIter::operator--() {
  pos--; return *this; }

////////////////////////////////////////////////////////////

inline
FFTData::CIter FFTData::CIter::operator--(int) {
  CIter tmp = *this; pos--; return tmp; }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                       BaseFFT

class BaseFFT {
  friend ostream& operator<<(ostream& os, const BaseFFT& fft) {
    return fft.printOn(os);
  }
  public:
    typedef FFTData Container;
    typedef Container::size_type size_type;
    enum WinFunc { NONE, HAMMING, HANNING };

    explicit BaseFFT(size_type winSize = 0, WinFunc winFunc = NONE);
    virtual ~BaseFFT();

    bool operator==(const BaseFFT& fft) const;
    bool operator!=(const BaseFFT& fft) const;

    void set(size_type winSize, WinFunc winFunc);

    size_type winSize() const { return _winSize; }
    size_type bits() const { return _bits; }
    WinFunc winFunc() const { return _winFunc; }

    void fft(Container* data) const;
    void invfft(Container* data) const;
    void fft(const AudioCache& wave, Container* spec) const;

  protected:
    virtual ostream& printOn(ostream& os) const;

  private:
    BaseFFT(const BaseFFT&) { }
    BaseFFT& operator=(const BaseFFT&) { return *this; }

    void fftAlgorithm(Container* data) const;
    void invfftAlgorithm(Container* data) const;
    void scale(Container* data) const;
    void window(Container* data) const;
    void unwindow(Container* data) const;
    const size_type reverse(size_type i) const;
    void reorder(Container* data) const;
    const size_type permute(size_type i) const { return _sigma[i]; }

    void copy(const AudioCache& wave, Container* spec) const;
    void deleteTables();
    void createTables();
    void setTables();

    size_type _bits;
    size_type _winSize;
    WinFunc _winFunc;
    Float* _sineTable;
    Float* _cosineTable;
    Float* _winTable;
    size_type* _sigma;
};

/////////////////////////////////////////////////////////////////

inline 
void BaseFFT::deleteTables() {
  delete[] _sineTable;
  delete[] _cosineTable;
  delete[] _winTable;
  delete[] _sigma;
}

/////////////////////////////////////////////////////////////////

inline
BaseFFT::~BaseFFT() { deleteTables(); }

/////////////////////////////////////////////////////////////////

inline
void BaseFFT::scale(Container* wave) const {
  for (size_type i = 0; i < _winSize; i++)
    wave->_data[i] = wave->_data[i] / _winSize;
}

/////////////////////////////////////////////////////////////////

inline
void BaseFFT::window(Container* wave) const {
  for (size_type i = 0; i < _winSize; i++)
    wave->_data[i] = wave->_data[i] * _winTable[i];
}

/////////////////////////////////////////////////////////////////

inline
void BaseFFT::unwindow(Container* wave) const {
  for (size_type i = 0; i < _winSize; i++)
    wave->_data[i] = wave->_data[i] / _winTable[i];
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
//                         FFT

typedef BaseFFT FFT;

#endif

