/*
   This file is part of the Aulib library.
   Name: rangeTableVectorType.cpp
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include <cstdio>
#include <iostream.h>
#include <strstream.h>
#include "rangeTableVectorType.H"


rangeTableVectorType::rangeTableVectorType(const rangeTableVectorType& t)
{
  vector<RangeTableVEType>::const_iterator iter;
  for (iter = t.rt.begin() ; iter<t.rt.end() ; iter++) {
    rt.push_back(*iter);
  }
}


rangeTableVectorType&
rangeTableVectorType::operator= (const rangeTableVectorType& t)
{
  if (this == &t)
    return *this;
  rt.erase(rt.begin(),rt.end());
  vector<RangeTableVEType>::const_iterator iter;
  for (iter = t.rt.begin() ; iter<t.rt.end() ; iter++) {
    rt.push_back(*iter);
  }
  return *this;
  }


// converts string into RangeTableVectorType
void
rangeTableVectorType::fromString (string s)
{
  // return immediately if string is too short
  if (s.length() < 6) return;
  istrstream inputString ((char*) s.c_str());
  string text;
  int from, to;
  while (!inputString.eof()) {
    inputString >> text ;
    // return if rest of string is empty
    if (inputString.eof()) break;
    inputString >> from >> to;
    rt.push_back(RangeTableVEType(text.c_str(),from, to));
  }
}

// converts RangeTableVectorType into string
string
rangeTableVectorType::toString ()
{
  vector<RangeTableVEType>::iterator iter;
  string s = "";
  char buffer1[10], buffer2[10];
  for (iter = rt.begin();iter<rt.end();iter++) {
    sprintf(buffer1,"%d",(*iter).start);
    sprintf(buffer2,"%d",(*iter).stop);
    s=s+(*iter).text + " " + buffer1 + " " + buffer2 + "\n";
  }
  return s;
}


// -----------------------------------------------------------------------
// query parts of the range table
// -----------------------------------------------------------------------
RangeTableVEType
rangeTableVectorType::getRange(int frameNo)
{
  vector<RangeTableVEType>::iterator iter;
  for (iter = begin() ; iter < end() ; iter++) {
    if ( ((*iter).start <= frameNo) && ((*iter).stop >= frameNo)) {
      return *iter;
    }
  }
  return RangeTableVEType();
}


// return range the frameNo falls into
rangeTableVectorType
rangeTableVectorType::getRange(int startFrameNo, int stopFrameNo)
{
  rangeTableVectorType newRt;
  vector<RangeTableVEType>::iterator iter;
  for (iter = begin() ; iter < end() ; iter++) {

    if ( ((*iter).start <= startFrameNo) && ((*iter).stop >= startFrameNo))
      newRt.rt.push_back(*iter);

    else if ( ((*iter).start >= startFrameNo) && ((*iter).stop <= stopFrameNo))
      newRt.rt.push_back(*iter);

    else if ( ((*iter).start <= stopFrameNo) && ((*iter).stop >= stopFrameNo))
      newRt.rt.push_back(*iter);

  }
  return newRt;
}

rangeTableVectorType
rangeTableVectorType::getRange(const RangeTableVEType& rte)
{
  getRange(rte.start, rte.stop);
}

// -----------------------------------------------------------------------
// remove individual ranges from the range table
// return 1 if something was removed otherwise 0
// -----------------------------------------------------------------------
bool
rangeTableVectorType::delRange(int frameNo)
{
  bool erased = false;
  vector<RangeTableVEType>::iterator iter;
  for (iter = begin() ; iter < end() ; iter++) {
    if ( ((*iter).start <= frameNo) && ((*iter).stop >= frameNo)) {
      rt.erase(iter);
      erased = true;
      iter--;
    }
  }
  return erased;
}


bool
rangeTableVectorType::delRange(int startFrameNo, int stopFrameNo)
{
  bool erased = false;
  vector<RangeTableVEType>::iterator iter, tempIter;

  for (iter = begin() ; iter < end() ; iter++) {
    if ( ((*iter).start <= startFrameNo) && ((*iter).stop >= startFrameNo)) {
      rt.erase(iter);
      erased = true;
      iter--;
    }

    else if ( ((*iter).start >= startFrameNo) && ((*iter).stop <= stopFrameNo)) {
      rt.erase(iter);
      erased = true;
      iter--;
    }

    else if ( ((*iter).start <= stopFrameNo) && ((*iter).stop >= stopFrameNo)) {
      erased = true;
      rt.erase(iter);
      iter--;
    }
  }
  return erased;
}

// included by S.P., deletes all ranges with rttext
bool
rangeTableVectorType::delRange(string rttext)
{
  bool erased = false;
  vector<RangeTableVEType>::iterator iter;

  for (iter = begin() ; iter < end() ; iter++) {
    if (!strncmp((const char*)((*iter).text.c_str()),
		 (const char*) (rttext.c_str()),
		 strlen((const char*)(rttext.c_str())) ) ) {
      rt.erase(iter);
      iter--;
      erased = true;
    }
  }
  return erased;
}

bool
rangeTableVectorType::delRange(const RangeTableVEType& rte)
{
  delRange(rte.start, rte.stop);
}


// -----------------------------------------------------------------------
// calculate total number of frames covered by the rangeTable
// -----------------------------------------------------------------------
int
rangeTableVectorType::noOfCoveredFrames() const
{
  int sum = 0;

  vector<RangeTableVEType>::const_iterator iter;
  // NUR "begin()" und "end()" geht nicht, das sie keinen
  // const-Zeiger zurueckgeben!!
  for (iter = rt.begin() ; iter < rt.end() ; iter++) {
    sum += (*iter).noOfCoveredFrames();
  }
  return sum;
}


void
rangeTableVectorType::add(rangeTableVectorType& rt2)
{
  vector<RangeTableVEType>::iterator iter;
  for (iter = rt2.begin() ; iter < rt2.end() ; iter++)
    rt.push_back(*iter);
}


ostream&
operator<< (ostream& os, const rangeTableVectorType& rt)
{
  vector<RangeTableVEType>::const_iterator iter;
  for (iter = rt.rt.begin() ; iter < rt.rt.end() ; iter++) {
    os << (*iter).text << "\t"
       << (*iter).start << "\t"
       << (*iter).stop << endl;
  }
  return os;
}

istream&
operator>>(istream& is , rangeTableVectorType& rt)
{
  string text;
  int from, to;

  while (!is.eof()) {
    is >> text ;
    // return if rest of string is empty
    // if (inputString.eof()) break;
    is >> from >> to;
    if (!is.eof()) {
      rt.rt.push_back(RangeTableVEType(text.c_str(),from, to));
    }
  }

  return is;
}
