/*
   This file is part of the Aulib library.
   Name: scalarFile.cpp
   Copyright (C) 2004 Silvia Pfeiffer

   Lehrstuhl fuer Praktische Informatik IV
   Universitaet Mannheim
   68131 Mannheim
   Germany
   email: pfeiffer@pi4.informatik.uni-mannheim.de

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/

#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <float.h>

#ifndef _WIN32
#ifdef DAMN_INTEL_BYTE_ORDER
#include <sys/mode.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#endif not _WIN32

#include "byteOrder.h"
#include "scalarFile.H"

// ---------------------------------------------------------------------
// -  CLASS scalarFile
// -  scalar element type specified by T
// ---------------------------------------------------------------------
int
scalarFile::newDB(string _filename,
		  int _startFrameNo, int _stopFrameNo)
{
  filename = _filename;
  startFrameNo = _startFrameNo;
  stopFrameNo = _stopFrameNo;

  if ( (file = fopen(filename.c_str(),"wb")) == NULL) {
    cerr << "Cannot create DB file "
	 << filename.c_str()
	 << "." << endl;
    return false;
  }

  // write header
  int temp_startFrameNo = startFrameNo,
    temp_stopFrameNo = stopFrameNo;
#ifdef DAMN_INTEL_BYTE_ORDER
  swap_int ((char*)&temp_startFrameNo);
  swap_int ((char*)&temp_stopFrameNo);
#endif
  fwrite (&temp_startFrameNo, sizeof(int), 1, file);
  fwrite (&temp_stopFrameNo, sizeof(int), 1, file);

  minValue=FLT_MAX;
  maxValue=-FLT_MAX;
  float temp_minValue = minValue,
    temp_maxValue = maxValue;
#ifdef DAMN_INTEL_BYTE_ORDER
  swap_int ((char*)&temp_minValue);
  swap_int ((char*)&temp_maxValue);
  fwrite ((char*)&temp_minValue, sizeof(unsigned char), 4, file);
  fwrite ((char*)&temp_maxValue, sizeof(unsigned char), 4, file);
#else
  fwrite (&temp_minValue, sizeof(float), 1, file);
  fwrite (&temp_maxValue, sizeof(float), 1, file);
#endif
//   fwrite (&size, sizeof(int), 1, file);

  // write data
  scalarType empty;
  for (int i=startFrameNo ; i<=stopFrameNo ; i++) {
    empty.write(file);
  }
  fclose(file);
#ifndef _WIN32
  chmod (filename.c_str(),S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
#endif not _WIN32

  // reopen file
  file = fopen(filename.c_str(),"r+b");

  return true;
}


int
scalarFile::openDB(string _filename,
		   int _startFrameNo, int _stopFrameNo,
		   bool newIfFailed)
{
  filename = _filename;
  startFrameNo = _startFrameNo;
  stopFrameNo = _stopFrameNo;

  if (file) {
    fclose(file);
    file = NULL;
  }
  file = fopen(filename.c_str(),"r+b");
  if (!file) {
    if (newIfFailed)
      return newDB(filename, startFrameNo, stopFrameNo);
    else
      return false;
  }

  // check, whether DB file has write size
  int dbstartFrameNo, dbstopFrameNo, dbsize;
  fread(&dbstartFrameNo, sizeof(int), 1, file);
  fread(&dbstopFrameNo, sizeof(int), 1, file);

#ifdef DAMN_INTEL_BYTE_ORDER
  fread((char*)&minValue, sizeof(unsigned char), 4, file);
  fread((char*)&maxValue, sizeof(unsigned char), 4, file);
  swap_int ((char*)&dbstartFrameNo);
  swap_int ((char*)&dbstopFrameNo);
  swap_int ((char*)&minValue);
  swap_int ((char*)&maxValue);
#else
  fread(&minValue, sizeof(float), 1, file);
  fread(&maxValue, sizeof(float), 1, file);
#endif


//   fread(&dbsize, sizeof(int), 1, file);

  if ( (dbstartFrameNo != startFrameNo) ||
       (dbstopFrameNo != stopFrameNo) ) {
    // DB has wrong size, generate new
    if (newIfFailed)
      return newDB(filename, startFrameNo, stopFrameNo);
    else
      return false;
  }

  return true;
}


int
scalarFile::save(int frameNo, scalarType value)
{
  if (! file) return false;
  // write result to file
  fseek(file,
	(frameNo - startFrameNo) * value.fileSize() +
	2 * sizeof(int) + 2 * sizeof(float),
// 	3 * sizeof(int) + 2 * sizeof(float),
	SEEK_SET);
  value.calculated = 1;
  value.write(file);

  // Update Min/Max Values in DB
  if (value.value < minValue) {
    minValue = value.value;
    fseek(file,
	  2 * sizeof(int),
	  SEEK_SET);
    float temp_minValue = minValue;
#ifdef DAMN_INTEL_BYTE_ORDER
    swap_int ((char*)&temp_minValue);
    fwrite((char*)&temp_minValue, sizeof(unsigned char), 4, file);
#else
    fwrite(&temp_minValue, sizeof(float), 1, file);
#endif
  }

  if (value.value > maxValue) {
    maxValue = value.value;
    fseek(file,
	  2 * sizeof(int) + sizeof(float),
	  SEEK_SET);
  float temp_maxValue = maxValue;
#ifdef DAMN_INTEL_BYTE_ORDER
    swap_int ((char*)&temp_maxValue);
    fwrite((char*)&temp_maxValue, sizeof(unsigned char), 4, file);
#else
    fwrite(&temp_maxValue, sizeof(float), 1, file);
#endif
  }

  return true;
}


scalarType
scalarFile::load(int frameNo)
{
  scalarType value (0);
  if (! file) return value;
  // read from file
  fseek(file,
	(frameNo - startFrameNo) * value.fileSize() +
	2 * sizeof(int) + 2 * sizeof(float),
	SEEK_SET);
  value.read(file);
  return value;
}

scalarFile::scalarFile () {
  file = NULL;
  return;
}
scalarFile::~scalarFile () {
  if (file)
    fclose (file);
  return;
}
